import json
import shutil
import subprocess
from abc import ABCMeta, abstractmethod
from functools import cache
from threading import Lock
from typing import Optional

from ..state import VERSIONS, Version
from ..utils import *
from .base import Package


class Installer(metaclass=ABCMeta):
    @abstractmethod
    def install(self, package: str) -> None:
        pass

    @abstractmethod
    def upgrade(self, package: str) -> None:
        pass

    def is_installed(self, package: str) -> bool:
        return self.installed_version(package) is not None

    @abstractmethod
    def installed_version(self, package: str) -> Optional[str]:
        pass

    @abstractmethod
    def latest_version(self, package: str) -> str:
        pass


class Brew(Installer):
    def install(self, package: str) -> None:
        run("brew", "install", package)

    def upgrade(self, package: str) -> None:
        run("brew", "upgrade", package)

    @cache
    def info(self, package: str) -> tuple[str, Optional[str]]:
        info = json.loads(run_output("brew", "info", "--json=v2", package))
        if info["casks"]:
            cask = info["casks"][0]
            return cask["version"], cask["installed"]
        if info["formulae"]:
            formula = info["formulae"][0]
            try:
                installed = formula["installed"][0]["version"]
            except IndexError:
                installed = None
            version = formula["versions"]["stable"]
            revision = formula.get("revision")
            if revision:
                version += f"_{revision}"
            return version, installed
        raise ValueError(f"Unexpected output from brew: {info}")

    def installed_version(self, package: str) -> Optional[str]:
        try:
            return self.info(package)[1]
        except IndexError:
            return None

    def latest_version(self, package: str) -> str:
        return self.info(package)[0]


class DNF(Installer):
    def install(self, package: str) -> None:
        run("sudo", "dnf", "install", "-y", package)

    def upgrade(self, package: str) -> None:
        run("sudo", "dnf", "upgrade", "-y", package)

    def installed_version(self, package: str) -> Optional[str]:
        try:
            output = run_output(
                "rpm",
                "--query",
                "--queryformat",
                "%{VERSION}",
                "--whatprovides",
                package,
                stderr=subprocess.DEVNULL,
            ).strip()
        except subprocess.CalledProcessError:
            return None
        if not output:
            return None
        return output

    def latest_version(self, package: str) -> str:
        output = run_output(
            "dnf",
            "--quiet",
            "repoquery",
            "--queryformat",
            "%{VERSION}",
            "--latest-limit",
            "1",
            "--arch",
            "x86_64,noarch",
            "--whatprovides",
            package,
        ).strip()
        if not output or "\n" in output:
            raise Exception(f"Cannot determine version for {package}.")
        return output


class Apt(Installer):
    def install(self, package: str) -> None:
        run("sudo", "apt", "install", "--yes", package)

    def upgrade(self, package: str) -> None:
        self.install(package)

    def latest_version(self, package: str) -> str:
        output = run_output(
            "apt-cache", "show", "--quiet", "--no-all-versions", package
        ).strip()
        for line in output.splitlines():
            line = line.strip()
            if line.startswith("Version:"):
                return line.split(": ", 1)[-1]
        raise Exception(f"Cannot determine version for {package}.")

    def installed_version(self, package: str) -> Optional[str]:
        try:
            return run_output(
                "dpkg-query",
                "--showformat",
                "${Version}",
                "--show",
                package,
                stderr=subprocess.DEVNULL,
            ).strip()
        except subprocess.CalledProcessError:
            return None


def linux_installer() -> Installer:
    if shutil.which("dnf"):
        return DNF()
    elif shutil.which("apt"):
        return Apt()
    else:
        raise NotImplementedError("Cannot find a package manager.")


INSTALLER: Installer = with_os(
    linux=linux_installer, macos=lambda: Brew()  # pylint:disable=unnecessary-lambda
)()

INSTALLER_LOCK = Lock()


class SystemPackage(Package):
    def __init__(
        self,
        *,
        name: str,
        url: Optional[str] = None,
        auto_updates: bool = False,
        service: Some[str] = None,
        **kwargs,
    ) -> None:
        self._name = name
        self.url = url
        self.auto_updates = auto_updates
        self.services = unsome(service)
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return self._name

    def get_remote_version(self) -> str:
        if self.url:
            return url_version(self.url)
        if self.auto_updates:
            return "latest"
        return INSTALLER.latest_version(self.name)

    @property
    def local_version(self) -> Optional[str]:
        if self.url:
            try:
                return VERSIONS[self.name].version
            except KeyError:
                return None
        if self.auto_updates:
            return "latest" if INSTALLER.is_installed(self.name) else None
        return INSTALLER.installed_version(self.name)

    def postinstall_linux(self):
        if self.services:
            run("sudo", "systemctl", "daemon-reload")
            for service in self.services:
                run("sudo", "systemctl", "enable", service)

    def postinstall_macos(self):
        pass

    def install(self):
        with INSTALLER_LOCK:
            if self.url:
                INSTALLER.install(self.url)
                VERSIONS[self.name] = Version(version=self.get_remote_version())
            elif self.local_version:
                INSTALLER.upgrade(self.name)
            else:
                INSTALLER.install(self.name)
        with_os(linux=self.postinstall_linux, macos=self.postinstall_macos)()



import json
import shutil
from dataclasses import dataclass
from functools import cache
from typing import Any, Callable, Iterable

import requests

from ..utils import *
from .archive import ArchivePackage


@cache
def have_github_auth() -> bool:
    return bool(shutil.which("gh")) and run_ok("gh", "auth", "status")


def github_api(url: str) -> Any:
    if have_github_auth():
        return json.loads(run_output("gh", "api", url))
    else:
        return requests.get(f"https://api.github.com/{url}").json()


@dataclass
class GitHubReleaseArtifact:
    name: str
    url: str


@dataclass
class GitHubRelease:
    tag_name: str
    assets: list[GitHubReleaseArtifact]


class GitHubPackage(ArchivePackage):
    prefixes: list[str]
    suffixes: list[str]
    excludes: list[str]

    def __init__(
        self,
        *,
        repo: str,
        prefix: Some[str] = None,
        suffix: Some[str] = None,
        exclude: Some[str] = None,
        **kwargs,
    ) -> None:
        self.repo = repo
        self.prefixes = unsome(prefix)
        self.suffixes = unsome(suffix)
        self.excludes = unsome(exclude)
        super().__init__(**kwargs)

    @cache
    def latest_release(self) -> GitHubRelease:
        latest = github_api(f"repos/{self.repo}/releases/latest")
        return GitHubRelease(
            tag_name=latest["tag_name"],
            assets=[
                GitHubReleaseArtifact(
                    name=result["name"], url=result["browser_download_url"]
                )
                for result in latest["assets"]
            ],
        )

    def filters(self) -> Iterable[Callable[[str], bool]]:
        for prefix in self.prefixes:
            yield Filters.startswith(prefix)
        for suffix in self.suffixes:
            yield Filters.endswith(suffix)
        for exclude in self.excludes:
            yield Filters.excludes(exclude)
        for hint in [".tar.gz"]:
            yield Filters.includes(hint)
        os_hints = with_os(
            linux=["linux", "gnu"],
            macos=["macos", "darwin", "osx"],
        )
        for hint in os_hints:
            yield Filters.includes(hint)
        arch_hints = ["x86_64", "amd64"]
        for hint in arch_hints:
            yield Filters.includes(hint)

    def artifact(self) -> GitHubReleaseArtifact:
        candidates = self.latest_release().assets
        for filter_fn in self.filters():
            if len(candidates) == 1:
                return candidates[0]
            new_candidates = [
                candidate for candidate in candidates if filter_fn(candidate.name)
            ]
            if new_candidates:
                candidates = new_candidates
            if len(candidates) == 1:
                return candidates[0]
        raise ValueError(f"Cannot choose between: {candidates}")

    def archive_url(self) -> str:
        return self.artifact().url

    @property
    def name(self) -> str:
        return self.repo

    def get_remote_version(self) -> str:
        return self.latest_release().tag_name



from pathlib import Path
from typing import Optional

from ..fs import HOME
from ..utils import *
from .base import Package


class Clone(Package):
    def __init__(self, clone: str, destination: str, **kwargs) -> None:
        self.repo = clone
        self.destination = destination
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return self.repo

    @property
    def directory(self) -> Path:
        return HOME / self.destination

    @property
    def directory_exists(self) -> bool:
        return self.directory.is_dir()

    @property
    def local_version(self) -> Optional[str]:
        if not self.directory_exists:
            return None
        return run_output("git", "rev-parse", "HEAD", cwd=self.directory)

    @property
    def remote(self):
        return f"https://github.com/{self.repo}.git"

    def get_remote_version(self) -> str:
        return run_output("git", "ls-remote", self.remote, "HEAD").split()[0]

    def install(self) -> None:
        if not self.directory_exists:
            run("git", "clone", self.remote, str(self.directory))
        run("git", "remote", "set-url", "origin", self.remote, cwd=self.directory)
        run("git", "fetch", cwd=self.directory)
        default_branch = run_output(
            "git", "rev-parse", "--abbrev-ref", "origin/HEAD", cwd=self.directory
        ).split("/")[1]
        run("git", "switch", default_branch, cwd=self.directory)
        run("git", "reset", "--hard", f"origin/{default_branch}", cwd=self.directory)



import configparser
import hashlib
import tempfile

from ..utils import *
from .manual_version import ManualVersion


class YumRepo(ManualVersion):
    def __init__(
        self, yum_name: str, yum_url: str, gpg_key: Optional[str] = None, **kwargs
    ) -> None:
        self.repo_name = yum_name
        self.baseurl = yum_url
        self.gpg_key = gpg_key
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return f"yum-{self.repo_name}"

    def get_remote_version(self) -> str:
        m = hashlib.sha256()
        m.update(self.baseurl.encode())
        m.update((self.gpg_key or "").encode())
        return m.hexdigest()

    def install(self) -> None:
        if CURRENT_OS != "linux":
            raise Exception("YumRepo is only supported on Linux")
        if CURRENT_DISTRIBUTION != "fedora":
            raise Exception("YumRepo is only supported on Fedora")

        repo = configparser.ConfigParser()
        repo[self.repo_name] = {
            "name": self.repo_name,
            "baseurl": self.baseurl,
            "enabled": "1",
        }
        if self.gpg_key is not None:
            repo[self.repo_name]["gpgcheck"] = "1"
            repo[self.repo_name]["gpgkey"] = self.gpg_key

        with tempfile.NamedTemporaryFile(mode="w") as repo_file:
            repo.write(repo_file, space_around_delimiters=False)
            repo_file.flush()
            run(
                "sudo",
                "install",
                "--mode",
                "644",
                repo_file.name,
                f"/etc/yum.repos.d/{self.repo_name}.repo",
            )

        if self.gpg_key:
            run("sudo", "rpm", "--import", self.gpg_key)

        super().install()



import os
import shutil
from pathlib import Path
from typing import Literal, Optional

from .utils import run


def find_executable(*executables: str) -> str:
    for candidate in executables:
        if shutil.which(candidate):
            return candidate
    raise Exception(f"None of {','.join(executables)} found in PATH.")


def is_executable(path: Path) -> bool:
    return os.access(path, os.X_OK)


LN = find_executable("gln", "ln")


HOME = Path.home()

LOCAL = HOME / ".local"


def transplant_path(dir_from: Path, dir_to: Path, path: Path) -> Path:
    return dir_to.joinpath(path.relative_to(dir_from))


def makedirs(path: Path, sudo: bool = False) -> None:
    if sudo:
        run("sudo", "mkdir", "-p", str(path))
    else:
        path.mkdir(parents=True, exist_ok=True)


def rm(path: Path, sudo: bool = False) -> None:
    if sudo:
        run("sudo", "rm", "-r", "-f", str(path))
    else:
        path.unlink(missing_ok=True)


def make_executable(path: Path) -> None:
    run("chmod", "+x", str(path))


LinkMethod = Literal["binary_wrapper"]


def link(
    source: Path,
    target: Path,
    *,
    sudo: bool = False,
    method: Optional[LinkMethod] = None,
) -> None:
    makedirs(target.parent, sudo=sudo)
    rm(target, sudo=sudo)
    if method == "binary_wrapper":
        if sudo:
            raise NotImplementedError()
        else:
            with open(target, "w") as wrapper_file:
                print(f'#!/bin/sh\nexec "{source}" "$@"', file=wrapper_file)
            make_executable(target)
    else:
        if sudo:
            run("sudo", LN, "-s", "-f", "-T", str(source), str(target))
        else:
            target.symlink_to(source)



import hashlib
from pathlib import Path
from typing import Iterator

from ..fs import HOME, link
from ..utils import *
from .manual_version import ManualVersion


class Links(ManualVersion):
    dest: Path

    def __init__(
        self,
        links: str,
        dest: str,
        dot: bool = False,
        root: bool = False,
        shallow: bool = False,
        only: Some[str] = None,
        **kwargs,
    ) -> None:
        self.root = root
        self.source = Path(links).absolute()
        if dest.startswith("/"):
            self.dest = Path(dest)
        else:
            if self.root:
                self.dest = Path("~root").expanduser() / dest
            else:
                self.dest = HOME / dest
        self.dot = dot
        self.shallow = shallow
        self.only = unsome_(only)
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return f"links-{self.source}-{self.dest}"

    def all_paths(self) -> Iterator[Path]:
        if self.shallow:
            candidates = self.source.glob("*")
        else:
            candidates = self.source.rglob("*")

        for path in candidates:
            if path.is_file():
                yield path

    def paths(self) -> Iterator[Path]:
        for path in self.all_paths():
            if self.only and path.name not in self.only:
                continue
            yield path

    def get_remote_version(self) -> str:
        m = hashlib.sha256()
        for path in self.paths():
            m.update(str(path).encode())
        return m.hexdigest()

    def install(self) -> None:
        # TODO: remove links that were created before but no longer exist
        for path in self.paths():
            target = path.relative_to(self.source)
            if self.dot:
                first_part, *parts = target.parts
                target = Path("." + first_part, *parts)
            target = self.dest.joinpath(target)

            link(path, target, sudo=self.root)
        super().install()



import argparse

from .package import load_packages
from .utils import flatten, parallel_map_tqdm


def parser() -> argparse.ArgumentParser:
    result = argparse.ArgumentParser(description="Install the environment")
    result.add_argument("component", nargs="*", help="The component to install")
    result.add_argument(
        "--sequential", action="store_true", help="Install packages one at a time"
    )
    return result


def main():
    args = parser().parse_args()
    components: frozenset[str] = frozenset(args.component) | {"base"}
    packages = flatten(map(load_packages, components))
    map_fn = map if args.sequential else parallel_map_tqdm
    results = map_fn(lambda p: p.ensure(), packages)
    total_installed = sum(1 if installed else 0 for installed in results)
    print(f"{total_installed} packages installed or updated.")



import os
import pwd
from pathlib import Path
from typing import Optional

from ..fs import is_executable
from ..utils import run
from .base import Package

SHELLS_FILE = "/etc/shells"


def get_all_shells() -> list[str]:
    with open(SHELLS_FILE) as shells_file:
        return [shell.strip() for shell in shells_file]


class Shell(Package):
    def __init__(self, shell: str, **kwargs) -> None:
        self.shell = Path(shell)
        super().__init__(**kwargs)

    @property
    def name(self) -> str:
        return "_shell"

    def get_remote_version(self) -> str:
        return str(self.shell)

    @property
    def local_version(self) -> Optional[str]:
        return pwd.getpwuid(os.getuid()).pw_shell

    def install(self) -> None:
        if not self.shell.is_file():
            raise ValueError(f"{self.shell} does not exist.")
        if not is_executable(self.shell):
            raise ValueError(f"{self.shell} is not executable.")
        if str(self.shell) not in get_all_shells():
            run("sudo", "tee", "-a", SHELLS_FILE, input=str(self.shell).encode())
        run("chsh", "-s", str(self.shell))
